lappend auto_path . ./win64/twapi
package require Tcl 8.6
package require TclOO
package require Tk
package require twapi

source "csv.tcl"

oo::class create Interval {
    variable begin
    variable end
    variable duration

    constructor {first {last 0}} {
        if {$last == 0} {
            set begin 0
            set end $first
        } else {
            set begin $first
            set end $last
        }
        set duration [expr {$end - $begin}]
    }

    method add {amount} {
        incr duration $amount
    }

    method format {} {
        set remainder $duration
        set hours [expr {int($remainder / 3600)}]
        set remainder [expr {$remainder - ($hours * 3600)}]
        set minutes [expr {int($remainder / 60)}]
        set remainder [expr {$remainder - ($minutes * 60)}]
        set seconds $remainder
        set retval [format "%02d:%02d:%02d" $hours $minutes $seconds]
        return $retval
    }

    method getDuration {} {
        return $duration
    }
}

oo::class create Timelord {
    variable history
    variable isClockOn
    variable labor
    variable seconds
    variable tray
    variable window

    constructor {args} {
        set seconds [clock seconds]
        set history [list]
        set labor 0
        set isClockOn false
        set window ""
        set icon [twapi::load_icon_from_file "tardis.ico"]
        set photo [image create photo -file "tardis.png"]
        ttk::button $window.export -command [list [self] export] -text "Export"
        ttk::button $window.start -command [list [self] start] -text "Start"
        ttk::button $window.stop -command [list [self] stop] -text "Stop"
        set row 0
        grid $window.export -column 0 -row $row -sticky ew
        grid $window.start -column 1 -row $row -sticky ew
        grid $window.stop -column 2 -row $row -sticky ew
        wm iconphoto . $photo
        wm protocol . WM_DELETE_WINDOW exit
        twapi::register_hotkey 107 [list $window.start invoke]
        twapi::register_hotkey 109 [list $window.stop invoke]
        set tray [twapi::systemtray addicon $icon [list [self] checkTray]]  
        rename ::exit ::exit.orig
        rename ::exit.new ::exit
    }

    method checkTray {id event location time} {
        if {$event eq "lbuttondown"} {
            set state [wm state .]
            if {$state eq "withdrawn"} {
                wm deiconify .
            } else {
                wm withdraw .
            }
        }
    }

    method exit {} {
        twapi::systemtray removeicon $tray
    }

    method export {} {
        [self] stop
        set now [clock seconds]
        set file [clock format $now -format "%Y%m%d-%H%M%S.csv"]
        set path [tk_getSaveFile                        \
            -defaultextension ".csv"                    \
            -filetypes [list [list "CSV Files" ".csv"]] \
            -initialdir "~"                             \
            -initialfile $file                          \
            -title "Choose export file name..."         \
        ]
        if {[string length $path] > 0} {
            set data [$::csv formatDicts $history]
            set fh [open $path "w"]
            puts $fh $data
            close $fh
            exec {*}[auto_execok start] "" $path &
        }
    }

    method getClockStr {} {
        if {$isClockOn} {
            set retval "on"
        } else {
            set retval "off"
        }
        return $retval
    }

    method getClockLaborStr {} {
        set clockStr [[self] getClockStr]
        set laborStr [[self] getLaborStr]
        set retval "$clockStr $laborStr"
        return $retval
    }

    method getLaborStr {} {
        if {$isClockOn} {
            set now [clock seconds]
            set interval [Interval new $seconds $now]
            $interval add $labor
        } else {
            set interval [Interval new $labor]
        }
        set retval [$interval format]
        return $retval
    }

    method go {} {
        [self] update
    }

    method keypress {keycode} {
        switch $keycode {
        107 {
            $window.start invoke
        }
        109 {
            $window.stop invoke
        }
        }
    }

    method log {} {
        set now [clock seconds]
        set time [clock format $now]
        set laborStr [[self] getLaborStr]

        set data [dict create    \
            time $time           \
            isClockOn $isClockOn \
            now $now             \
            labor $labor         \
            laborStr $laborStr   \
        ]
        lappend history $data

        set text [[self] getClockLaborStr]
        twapi::systemtray modifyicon $tray -balloon $text
    }

    method refresh {} {
        [self] update
        if {$isClockOn} {
            after 1 [list [self] refresh]
        }
    }

    method start {} {
        if {!$isClockOn} {
            set seconds [clock seconds]
            set isClockOn true
            [self] log
            [self] refresh
        }
    }

    method stop {} {
        if {$isClockOn} {
            set now [clock seconds]
            set interval [Interval new $seconds $now]
            set isClockOn false
            incr labor [$interval getDuration]
            [self] log
        }
    }

    method update {} {
        set text [[self] getClockLaborStr]
        wm title . $text
    }
}

proc ::exit.new {} {
    $::timelord exit
    exit.orig
}

set ::csv [Csv new]
set ::timelord [Timelord new]
$::timelord go
